/* Copyright (c) 2004-2011 dhpoware. All rights reserved.
 *
 * Permission is hereby granted, free of charge, to any person 
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without 
 * restriction, including without limitation the rights to use, 
 * copy, modify, merge, publish, distribute, sublicense, and/or 
 * sell copies of the Software, and to permit persons to whom 
 * the Software is furnished to do so, subject to the following 
 * conditions:
 *
 * The above copyright notice and this permission notice shall be 
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES 
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND 
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT 
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, 
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR 
 * OTHER DEALINGS IN THE SOFTWARE.
 */

/* This is a native Windows application used to generate textures
 * containing anti-aliased TrueType and OpenType font character sets.
 */

// 'type case' : pointer truncation from 'HBRUSH' to 'LONG'
#pragma warning (disable : 4311)

#include <windows.h>
#include <commdlg.h>
#include <stdio.h>
#include <string.h>
#include "resource.h"
#include "win32_dib.h"


#define DEFAULT_TEXTURE_SIZE    512
#define APP_WND_STYLE           (WS_POPUPWINDOW | WS_CAPTION | WS_MINIMIZEBOX)


DIB      g_bitmap;
LOGFONT  g_lf;
HFONT    g_hFont;
HBRUSH   g_hBackgroundBrush;
COLORREF g_fontColor = RGB(255, 255, 255);
COLORREF g_backgroundColor = RGB(0, 0, 0);
char     g_szFilename[MAX_PATH + 1];
BYTE     g_lfQuality = DEFAULT_QUALITY;
BOOL     g_bEnableAntiAliasing = TRUE;
UINT     g_textureSize = DEFAULT_TEXTURE_SIZE;
UINT     g_currTextureSizeItem = IDM_TEXTURE_512;
int      g_pointSize = DEFAULT_TEXTURE_SIZE / 16;
int      g_logPixelsY;
int      g_fontHeight;


const char *SelectFilename(HWND hWnd, const char *pszFilter, const char *pszExt)
{
    OPENFILENAME ofn = {0};

    ofn.lStructSize = sizeof(ofn);
    ofn.hwndOwner = hWnd;
    ofn.lpstrFilter = pszFilter;
    ofn.nFilterIndex = 1;
    ofn.lpstrFile = g_szFilename;
    ofn.nMaxFile = sizeof(g_szFilename);
    ofn.lpstrTitle = "Export Font";
    ofn.Flags = OFN_OVERWRITEPROMPT;

    if (!GetSaveFileName(&ofn))
        return NULL;

    if (!strstr(g_szFilename, pszExt))
        strcat(g_szFilename, pszExt);

    return g_szFilename;
}

void SaveFontWidthsBinary(DIB *pBitmap, HFONT hFont, FILE *pFile)
{
    if (pBitmap && hFont && pFile)
    {
        int i;
        HDC hDC;
        HFONT hPrevFont;
        BYTE width;
        SIZE charSize;
        char szString[2] = {0};

        hDC = pBitmap->dc;
        hPrevFont = SelectObject(hDC, hFont);

        for (i = 0; i < 256; ++i)
        {
            szString[0] = i;
            GetTextExtentPoint32(hDC, szString, 1, &charSize);
            width = (BYTE)charSize.cx;
            fwrite(&width, sizeof(width), 1, pFile);
        }

        SelectObject(hDC, hPrevFont);
    }
}

void SaveFontWidthsCHeaderFile(DIB *pBitmap, HFONT hFont, FILE *pFile)
{
    if (pBitmap && hFont && pFile)
    {
        int i, j, ch;
        HDC hDC;
        HFONT hPrevFont;
        BYTE width;
        SIZE charSize;
        char szString[2] = {0};

        hDC = pBitmap->dc;
        hPrevFont = SelectObject(hDC, hFont);
        ch = 0;

        for (i = 0; i < 16; ++i)
        {
            for (j = 0; j < 16; ++j)
            {
                szString[0] = ch++;
                GetTextExtentPoint32(hDC, szString, 1, &charSize);
                width = (BYTE)charSize.cx;

                if (ch != 256)
                    fprintf(pFile, "%d, ", width);
                else
                    fprintf(pFile, "%d", width);
            }

            fprintf(pFile, "\n");
        }

        SelectObject(hDC, hPrevFont);
    }
}

void SaveFontAlphaTexture(DIB *pBitmap, FILE *pFile)
{
    if (pBitmap && pFile)
    {
        int x, y, width, height, pitch;
        BYTE alpha;
        BYTE *pPixels, *pRow;

        width = pBitmap->width;
        height = pBitmap->height;
        pitch = pBitmap->pitch;
        pPixels = pBitmap->pPixels;

        for (y = 0; y < height; ++y)
        {
            pRow = pPixels + ((height - 1 - y) * pitch);

            for (x = 0; x < width; ++x)
            {
                alpha = (BYTE)pRow[x * 3];
                fwrite(&alpha, sizeof(alpha), 1, pFile);
            }
        }
    }
}

void SaveFontAlphaTextureAsCHeaderFile(DIB *pBitmap, FILE *pFile)
{
    if (pBitmap && pFile)
    {
        int x, y, width, height, pitch;
        BYTE alpha;
        BYTE *pPixels, *pRow;

        width = pBitmap->width;
        height = pBitmap->height;
        pitch = pBitmap->pitch;
        pPixels = pBitmap->pPixels;

        for (y = 0; y < height; ++y)
        {
            pRow = pPixels + ((height - 1 - y) * pitch);

            for (x = 0; x < width; ++x)
            {
                alpha = (BYTE)pRow[x * 3];

                if ((y == height - 1) && (x == width - 1))
                    fprintf(pFile, "%3d", alpha);
                else
                    fprintf(pFile, "%3d, ", alpha);
            }

            fprintf(pFile, "\n");
        }
    }
}

void ExportToFile(HWND hWnd)
{
    const char *pszFilename;
    const char *pszFilter = "Font Files (*.font)\0*.font\0";
    const char *pszExt = ".font";
    
    if (g_hFont)
    {
        *g_szFilename = '\0';

        if (pszFilename = SelectFilename(hWnd, pszFilter, pszExt))
        {
            FILE *pFile = fopen(pszFilename, "wb");

            if (pFile)
            {
                SaveFontWidthsBinary(&g_bitmap, g_hFont, pFile);
                fwrite(&g_fontHeight, 4, 1, pFile);
                fwrite(&g_textureSize, 4, 1, pFile);
                SaveFontAlphaTexture(&g_bitmap, pFile);
                fclose(pFile);
            }
        }
    }
    else
    {
        MessageBox(hWnd, "No font to export", "FontGen", MB_ICONINFORMATION);
    }
}

void ExportToCHeaderFile(HWND hWnd)
{
    const char *pszFilename;
    const char *pszFilter = "C Header Files (*.h)\0*.h\0";
    const char *pszExt = ".h";

    if (g_hFont)
    {
        *g_szFilename = '\0';

        if (pszFilename = SelectFilename(hWnd, pszFilter, pszExt))
        {
            FILE *pFile = fopen(pszFilename, "wb");
            
            if (pFile)
            {
                // Extract the file's basename.
                const char *pszBasename = strrchr(pszFilename, '\\');

                if (pszBasename)
                    ++pszBasename;
                else
                    pszBasename = pszFilename;

                // Start of file - inclusion guards.
                fprintf(pFile, "/* Generated by FontGen. */\n\n");
                fprintf(pFile, "#if !defined(%s)\n", pszBasename);
                fprintf(pFile, "#define %s\n\n", pszBasename);

                // Font header info.
                fprintf(pFile, "int g_fontHeight = %d;\n", g_fontHeight);
                fprintf(pFile, "int g_fontAlphaTextureSize = %d;\n\n", g_textureSize);
                
                // Font glyph widths.
                fprintf(pFile, "unsigned char g_fontWidths[256] = {\n");
                SaveFontWidthsCHeaderFile(&g_bitmap, g_hFont, pFile);
                fprintf(pFile, "};\n\n");

                // Font alpha texture pixels.
                fprintf(pFile, "unsigned char g_fontAlphaTexture[%d] = {\n",
                    g_textureSize * g_textureSize);
                SaveFontAlphaTextureAsCHeaderFile(&g_bitmap, pFile);
                fprintf(pFile, "};\n\n");

                // End of file - inclusion guards.
                fprintf(pFile, "#endif /* !defined(%s) */\n", pszBasename);
            }
        }
    }
    else
    {
        MessageBox(hWnd, "No font to export", "FontGen", MB_ICONINFORMATION);
    }
}

void ResizeWindow(HWND hWnd, int x, int y, int w, int h, BOOL bRepaint)
{
    RECT rc;
    BOOL bCenter = TRUE;

    if (x > -1 && y > -1)
    {
        bCenter = FALSE;
        GetWindowRect(hWnd, &rc);
        x = rc.left;
        y = rc.top;
    }

    SetRect(&rc, 0, 0, w, h);
    AdjustWindowRectEx(&rc, APP_WND_STYLE, TRUE, 0);
    w = rc.right - rc.left;
    h = rc.bottom - rc.top;
    
    if (bCenter)
    {
        x = (GetSystemMetrics(SM_CXSCREEN) - w) / 2;
        y = (GetSystemMetrics(SM_CYSCREEN) - h) / 2;
    }

    MoveWindow(hWnd, x, y, w, h, bRepaint);
}

void UpdateCaption(HWND hWnd)
{
    char szBuffer[256];

    sprintf(szBuffer, "FontGen [%d x %d]", g_textureSize, g_textureSize);
    SetWindowText(hWnd, szBuffer);
}

void EnableAntiAliasing(BOOL bEnable)
{
    g_bEnableAntiAliasing = bEnable;
    g_lfQuality = (bEnable) ? ANTIALIASED_QUALITY : DEFAULT_QUALITY;
}

void SelectColor(HWND hWnd, COLORREF *pColor)
{
    static CHOOSECOLOR cc;
    static COLORREF customColors[16];

    cc.lStructSize = sizeof(cc);
    cc.hwndOwner = hWnd;
    cc.hInstance = NULL;
    cc.lpCustColors = customColors;
    cc.Flags = CC_FULLOPEN;
    cc.lCustData = 0;
    cc.lpfnHook = NULL;
    cc.lpTemplateName = NULL;

    if (ChooseColor(&cc))
        *pColor = cc.rgbResult;
}

void SelectTextureSize(HMENU hMenu, UINT textureSizeItem)
{
    UINT newTextureSize;

    // Uncheck previously selected menu item.
    CheckMenuItem(hMenu, g_currTextureSizeItem, MF_UNCHECKED);

    // Check newly selected menu item.
    g_currTextureSizeItem = textureSizeItem;
    CheckMenuItem(hMenu, g_currTextureSizeItem, MF_CHECKED);

    // Determine the new texture size.
    switch (textureSizeItem)
    {
    default:
        newTextureSize = g_textureSize;
        break;

    case IDM_TEXTURE_128:
        newTextureSize = 128;
        break;

    case IDM_TEXTURE_256:
        newTextureSize = 256;
        break;

    case IDM_TEXTURE_512:
        newTextureSize = 512;
        break;

    case IDM_TEXTURE_1024:
        newTextureSize = 1024;
        break;

    case IDM_TEXTURE_2048:
        newTextureSize = 2048;
        break;
    }

    // Determine the actual texture size.
    if (newTextureSize != g_textureSize)
    {
        if (newTextureSize > g_textureSize)
        {
            // User wants a larger texture.
            g_pointSize *= newTextureSize / g_textureSize;
        }
        else
        {
            // User wants a smaller texture.
            g_pointSize /= g_textureSize / newTextureSize;
        }

        g_textureSize = newTextureSize;
    }

    DIBdestroy(&g_bitmap);
    DIBcreate(&g_bitmap, g_textureSize, g_textureSize);

    // Scale the font to fit inside the new texture size.
    if (g_lf.lfHeight)
    {
        HFONT hFont;

        // Select height of font based on chosen font's point size.
        //g_lf.lfHeight = -lfHeight;
        g_lf.lfHeight = -MulDiv(g_pointSize, g_logPixelsY, 72);

        // Force the font's smoothing quality. This is just a hint to
        // Windows - we may not get the required smoothing!
        g_lf.lfQuality = g_lfQuality;

        if (hFont = CreateFontIndirect(&g_lf))
        {
            if (g_hFont)
            {
                DeleteObject(g_hFont);
                g_hFont = NULL;
            }

            g_hFont = hFont;
        }
    }
}

BOOL SelectFont(HWND hWnd, BOOL bReload)
{
    // Let user select a font. The size of the font depends on the
    // currently selected texture size.

    static CHOOSEFONT cf;

    HFONT hNewFont;

    if (!bReload)
    {
        int ptSize;

        if (!g_logPixelsY)
        {
            HDC hDC;

            hDC = GetDC(hWnd);
            g_logPixelsY = GetDeviceCaps(hDC, LOGPIXELSY);
            ReleaseDC(hWnd, hDC);
        }

        ptSize = ((g_textureSize / 16) * 72) / g_logPixelsY;

        cf.lStructSize = sizeof(CHOOSEFONT);
        cf.hwndOwner = hWnd;
        cf.hDC = NULL;
        cf.lpLogFont = &g_lf;
        cf.iPointSize = 0;
        cf.Flags = CF_TTONLY | CF_SCREENFONTS | CF_LIMITSIZE
            | CF_NOFACESEL | CF_NOSIZESEL;
        cf.rgbColors = 0;
        cf.lCustData = 0;
        cf.lpfnHook = NULL;
        cf.lpTemplateName = NULL;
        cf.hInstance = NULL;
        cf.lpszStyle = NULL;
        cf.nFontType = SCREEN_FONTTYPE | REGULAR_FONTTYPE;
        cf.nSizeMin = 0;
        cf.nSizeMax = ptSize;
        
        if (!ChooseFont(&cf))
            return FALSE;

        g_pointSize = cf.iPointSize / 10;
    }
    
    // Select height of font based on chosen font's point size.
    g_lf.lfHeight = -MulDiv(g_pointSize, g_logPixelsY, 72);

    // Force the font's smoothing quality. This is just a hint to
    // Windows - we may not get the required smoothing!
    g_lf.lfQuality = g_lfQuality;
    
    hNewFont = CreateFontIndirect(&g_lf);

    if (!hNewFont)
        return FALSE;

    if (g_hFont)
    {
        DeleteObject(g_hFont);
        g_hFont = NULL;
    }

    g_hFont = hNewFont;
    return TRUE;
}

void DrawFontTexture(HFONT hFont, HDC hDC, HWND hWnd)
{
    // The font is layed out as a 16 x 16 grid of glyphs.
    // The window is resized as necessary to fit the texture.

    int i, j, x, y, ch, cellSize;
    RECT rc;
    TEXTMETRIC tm;
    COLORREF prevColor;
    HFONT hPrevFont;
    
    if (!hFont)
        return;

    // Save current settings and prepare for text rendering.
    hPrevFont = SelectObject(hDC, hFont);
    GetTextMetrics(hDC, &tm);
    prevColor = SetTextColor(hDC, g_fontColor);
    SetBkMode(hDC, TRANSPARENT);
    
    // Store the height of the font.
    g_fontHeight = tm.tmHeight;

    // Resize window to fit font character set.
    GetWindowRect(hWnd, &rc);
    ResizeWindow(hWnd, rc.left, rc.top, g_textureSize, g_textureSize, TRUE);
    GetClientRect(hWnd, &rc);
    UpdateCaption(hWnd);

    // Clear the texture.
    if (g_hBackgroundBrush)
        FillRect(hDC, &rc, g_hBackgroundBrush);
    else
        FillRect(hDC, &rc, (HBRUSH)GetStockObject(BLACK_BRUSH));

    // Render font character set.
    cellSize = g_textureSize / 16;
    ch = 0;
    
    for (i = 0; i < 16; ++i)
    {
        y = (cellSize * i) + (cellSize - tm.tmHeight);

        for (j = 0; j < 16; ++j)
        {
            x = cellSize * j + 1;
            TextOut(hDC, x, y, (LPCSTR)&ch, 1);            
            ++ch;
        }
    }

    // Restore previous settings.
    SetTextColor(hDC, prevColor);
    SelectObject(hDC, hPrevFont);
}

LRESULT CALLBACK WindowProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam)
{   
    HMENU hMenu;
    HDC hDC;
    PAINTSTRUCT ps;

    switch (msg)
    {
    default:
        break;

    case WM_CREATE:
        DIBcreate(&g_bitmap, g_textureSize, g_textureSize);
        return 0;

    case WM_COMMAND:
        switch (LOWORD(wParam))
        {
        default:
            break;

        case IDM_FILE_EXPORT:
            ExportToFile(hWnd);
            break;

        case IDM_FILE_EXPORT_AS_C_HEADER:
            ExportToCHeaderFile(hWnd);            
            break;

        case IDM_FILE_EXIT:
            PostMessage(hWnd, WM_CLOSE, 0, 0);
            break;

        case IDM_FONT_CHOOSE_FONT:
            if (SelectFont(hWnd, FALSE))
                InvalidateRect(hWnd, NULL, TRUE);
            break;

        case IDM_FONT_CHOOSE_FONT_COLOR:
            SelectColor(hWnd, &g_fontColor);
            InvalidateRect(hWnd, NULL, TRUE);
            break;

        case IDM_FONT_CHOOSE_BACKGROUND_COLOR:
            SelectColor(hWnd, &g_backgroundColor);
            if (g_hBackgroundBrush = CreateSolidBrush(g_backgroundColor))
            {
                SetClassLong(hWnd, GCL_HBRBACKGROUND, (LONG)g_hBackgroundBrush);
                InvalidateRect(hWnd, NULL, TRUE);
            }
            break;

        case IDM_FONT_ENABLE_ANTI_ALIASING:
            hMenu = GetMenu(hWnd);
            g_bEnableAntiAliasing ^= TRUE;
            if (g_bEnableAntiAliasing)
                CheckMenuItem(hMenu, IDM_FONT_ENABLE_ANTI_ALIASING, MF_CHECKED);
            else
                CheckMenuItem(hMenu, IDM_FONT_ENABLE_ANTI_ALIASING, MF_UNCHECKED);
            SelectFont(hWnd, TRUE);
            InvalidateRect(hWnd, NULL, TRUE);
            break;

        case IDM_TEXTURE_256:
        case IDM_TEXTURE_512:
        case IDM_TEXTURE_1024:
        case IDM_TEXTURE_2048:
            hMenu = GetMenu(hWnd);
            SelectTextureSize(hMenu, LOWORD(wParam));
            InvalidateRect(hWnd, NULL, TRUE);
            break;
        }
        return 0;

    case WM_DESTROY:
        if (g_hFont)
        {
            DeleteObject(g_hFont);
            g_hFont = NULL;
        }
        
        if (g_hBackgroundBrush)
        {
            DeleteObject(g_hBackgroundBrush);
            g_hBackgroundBrush = NULL;
        }

        DIBdestroy(&g_bitmap);
        PostQuitMessage(0);
        return 0;  

    case WM_PAINT:
        if (g_hFont && g_bitmap.dc)
        {
            if (hDC = BeginPaint(hWnd, &ps))
            {
                DIBselect(&g_bitmap);
                DrawFontTexture(g_hFont, g_bitmap.dc, hWnd);
                DIBblt(&g_bitmap, hDC, 0, 0);
                DIBdeselect(&g_bitmap);
                EndPaint(hWnd, &ps);
                return 0;
            }
        }
        break;
    }
    
    return DefWindowProc(hWnd, msg, wParam, lParam);
}

int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance,
                   LPSTR lpCmdLine, int nCmdShow)
{
    WNDCLASSEX wcl;
    HWND hWnd;
    MSG msg;
    
    wcl.cbSize = sizeof(wcl);
    wcl.style = CS_HREDRAW | CS_VREDRAW | CS_OWNDC;
    wcl.lpfnWndProc = WindowProc;
    wcl.cbClsExtra = 0;
    wcl.cbWndExtra = 0;
    wcl.hInstance = hInstance;
    wcl.hIcon = LoadIcon(NULL, IDI_APPLICATION);
    wcl.hCursor = LoadCursor(NULL, IDC_ARROW);
    wcl.hbrBackground = GetStockObject(BLACK_BRUSH);
    wcl.lpszMenuName = MAKEINTRESOURCE(IDR_MENU1);
    wcl.lpszClassName = "myWindowClass";
    wcl.hIconSm = NULL;
    
    if (!RegisterClassEx(&wcl))
        return 0;
    
    hWnd = CreateWindowEx(0, wcl.lpszClassName, "FontGen",
        APP_WND_STYLE, 0, 0, 0, 0, NULL, NULL, hInstance, NULL);
    
    if (hWnd)
    {
        ResizeWindow(hWnd, -1, -1, g_textureSize, g_textureSize, FALSE);
        ShowWindow(hWnd, nCmdShow);
        UpdateWindow(hWnd);
    
        while (GetMessage(&msg, NULL, 0, 0))
        {
            TranslateMessage(&msg);
            DispatchMessage(&msg);
        }
    }
    
    UnregisterClass(wcl.lpszClassName, wcl.hInstance);
    return 0;
}